@@ -1,4 +1,5 @@
 use std::collections::HashMap;
+use std::hash::{DefaultHasher, Hash, Hasher};
 use std::io::Read;
 use std::path::{Path, PathBuf};
 use std::sync::atomic::{AtomicBool, Ordering};
@@ -147,20 +148,15 @@ fn build_def_file(
         // Create the .def output file
         let def_file = cargo_util::paths::create(targetdir.join(format!("{name}.def")))?;
 
-        write_def_file(name, dll_file, def_file)?;
+        write_def_file(dll_file, def_file)?;
     }
 
     Ok(())
 }
 
-fn write_def_file<W: std::io::Write>(
-    name: &str,
-    dll_file: object::File,
-    mut def_file: W,
-) -> anyhow::Result<W> {
+fn write_def_file<W: std::io::Write>(dll_file: object::File, mut def_file: W) -> anyhow::Result<W> {
     use object::read::Object;
 
-    writeln!(def_file, "LIBRARY \"{name}.dll\"")?;
     writeln!(def_file, "EXPORTS")?;
 
     for export in dll_file.exports()? {
@@ -171,9 +167,10 @@ fn write_def_file<W: std::io::Write>(
     Ok(def_file)
 }
 
-/// Build import library for windows-gnu
+/// Build import library for windows
 fn build_implib_file(
     ws: &Workspace,
+    build_targets: &BuildTargets,
     name: &str,
     target: &target::Target,
     targetdir: &Path,
@@ -201,13 +198,15 @@ fn build_implib_file(
             }
         };
 
-        let lib_path = match flavor {
-            Flavor::Msvc => targetdir.join(format!("{name}.dll.lib")),
-            Flavor::Gnu => targetdir.join(format!("{name}.dll.a")),
-        };
+        let lib_name = build_targets
+            .shared_output_file_name()
+            .unwrap()
+            .into_string()
+            .unwrap();
+        let implib_path = build_targets.impl_lib.as_ref().unwrap();
 
-        let lib_file = cargo_util::paths::create(lib_path)?;
-        write_implib(lib_file, machine_type, flavor, &def_contents)?;
+        let implib_file = cargo_util::paths::create(implib_path)?;
+        write_implib(implib_file, lib_name, machine_type, flavor, &def_contents)?;
     }
 
     Ok(())
@@ -215,11 +214,14 @@ fn build_implib_file(
 
 fn write_implib<W: std::io::Write + std::io::Seek>(
     mut w: W,
+    lib_name: String,
     machine_type: MachineType,
     flavor: Flavor,
     def_contents: &str,
 ) -> anyhow::Result<W> {
-    let module_def = ModuleDef::parse(def_contents, machine_type)?;
+    let mut module_def = ModuleDef::parse(def_contents, machine_type)?;
+    module_def.import_name = lib_name;
+
     let import_library = ImportLibrary::from_def(module_def, machine_type, flavor);
 
     import_library.write_to(&mut w)?;
@@ -234,6 +236,7 @@ struct FingerPrint {
     build_targets: BuildTargets,
     install_paths: InstallPaths,
     static_libs: String,
+    hasher: DefaultHasher,
 }
 
 #[derive(serde::Serialize, serde::Deserialize)]
@@ -248,21 +251,24 @@ impl FingerPrint {
         root_output: &Path,
         build_targets: &BuildTargets,
         install_paths: &InstallPaths,
+        capi_config: &CApiConfig,
     ) -> Self {
+        let mut hasher = DefaultHasher::new();
+
+        capi_config.hash(&mut hasher);
+
         Self {
             id: id.to_owned(),
             root_output: root_output.to_owned(),
             build_targets: build_targets.clone(),
             install_paths: install_paths.clone(),
             static_libs: String::new(),
+            hasher,
         }
     }
 
     fn hash(&self) -> anyhow::Result<Option<String>> {
-        use std::collections::hash_map::DefaultHasher;
-        use std::hash::{Hash, Hasher};
-
-        let mut hasher = DefaultHasher::new();
+        let mut hasher = self.hasher.clone();
         self.install_paths.hash(&mut hasher);
 
         let mut paths: Vec<&PathBuf> = Vec::new();
@@ -323,7 +329,7 @@ impl FingerPrint {
     }
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub struct CApiConfig {
     pub header: HeaderCApiConfig,
     pub pkg_config: PkgConfigCApiConfig,
@@ -331,7 +337,7 @@ pub struct CApiConfig {
     pub install: InstallCApiConfig,
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub struct HeaderCApiConfig {
     pub name: String,
     pub subdirectory: String,
@@ -339,7 +345,7 @@ pub struct HeaderCApiConfig {
     pub enabled: bool,
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub struct PkgConfigCApiConfig {
     pub name: String,
     pub filename: String,
@@ -350,14 +356,14 @@ pub struct PkgConfigCApiConfig {
     pub strip_include_path_components: usize,
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub enum VersionSuffix {
     Major,
     MajorMinor,
     MajorMinorPatch,
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub struct LibraryCApiConfig {
     pub name: String,
     pub version: Version,
@@ -387,19 +393,19 @@ impl LibraryCApiConfig {
     }
 }
 
-#[derive(Debug, Default)]
+#[derive(Debug, Default, Hash)]
 pub struct InstallCApiConfig {
     pub include: Vec<InstallTarget>,
     pub data: Vec<InstallTarget>,
 }
 
-#[derive(Debug)]
+#[derive(Debug, Hash)]
 pub enum InstallTarget {
     Asset(InstallTargetPaths),
     Generated(InstallTargetPaths),
 }
 
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, Hash)]
 pub struct InstallTargetPaths {
     /// pattern to feed to glob::glob()
     ///
@@ -979,7 +985,13 @@ impl CPackage {
             args.get_flag("meson"),
         )?;
 
-        let finger_print = FingerPrint::new(&id, root_output, &build_targets, &install_paths);
+        let finger_print = FingerPrint::new(
+            &id,
+            root_output,
+            &build_targets,
+            &install_paths,
+            &capi_config,
+        );
 
         Ok(CPackage {
             version,
@@ -1021,7 +1033,7 @@ impl LibraryTypes {
         // - https://github.com/lu-zero/cargo-c/issues/180
         Self {
             staticlib: true,
-            cdylib: target.os != "none" && target.env != "musl",
+            cdylib: target.os != "none",
         }
     }
 
@@ -1133,7 +1145,7 @@ pub fn cbuild(
     for m in ws.members_mut().filter(|p| is_relevant_package(p)) {
         let cpkg = CPackage::from_package(m, args, library_types, &rustc_target, &root_output)?;
 
-        pristine |= cpkg.finger_print.load_previous().is_err();
+        pristine |= cpkg.finger_print.load_previous().is_err() || !cpkg.finger_print.is_valid();
 
         members.push(cpkg);
     }
@@ -1176,8 +1188,7 @@ pub fn cbuild(
 
     for cpkg in members.iter_mut() {
         // it is a new build, build the additional files and update update the cache
-        // if the hash value does not match.
-        if new_build && !cpkg.finger_print.is_valid() {
+        if new_build {
             let name = &cpkg.capi_config.library.name;
             let (pkg_config_static_libs, static_libs) = if library_types.only_cdylib() {
                 (String::new(), String::new())
@@ -1200,7 +1211,7 @@ pub fn cbuild(
             if !library_types.only_staticlib() && capi_config.library.import_library {
                 let lib_name = name;
                 build_def_file(ws, lib_name, &rustc_target, &root_output)?;
-                build_implib_file(ws, lib_name, &rustc_target, &root_output)?;
+                build_implib_file(ws, build_targets, lib_name, &rustc_target, &root_output)?;
             }
 
             if capi_config.header.enabled {
@@ -1387,11 +1398,13 @@ mod tests {
     pub fn test_lib_listing() {
         let libs_osx = "-lSystem -lc -lm";
         let libs_linux = "-lgcc_s -lutil -lrt -lpthread -lm -ldl -lc";
+        let libs_hurd = "-lgcc_s -lutil -lrt -lpthread -lm -ldl -lc";
         let libs_msvc = "kernel32.lib advapi32.lib kernel32.lib ntdll.lib userenv.lib ws2_32.lib kernel32.lib ws2_32.lib kernel32.lib msvcrt.lib /defaultlib:msvcrt";
         let libs_mingw = "-lkernel32 -ladvapi32 -lkernel32 -lntdll -luserenv -lws2_32 -lkernel32 -lws2_32 -lkernel32";
 
         let target_osx = target::Target::new(Some("x86_64-apple-darwin"), false).unwrap();
         let target_linux = target::Target::new(Some("x86_64-unknown-linux-gnu"), false).unwrap();
+        let target_hurd = target::Target::new(Some("x86_64-unknown-hurd-gnu"), false).unwrap();
         let target_msvc = target::Target::new(Some("x86_64-pc-windows-msvc"), false).unwrap();
         let target_mingw = target::Target::new(Some("x86_64-pc-windows-gnu"), false).unwrap();
 
@@ -1400,6 +1413,10 @@ mod tests {
             static_libraries(libs_linux, &target_linux),
             "-lgcc_s -lutil -lrt -lpthread -lm -ldl -lc"
         );
+        assert_eq!(
+            static_libraries(libs_hurd, &target_hurd),
+            "-lgcc_s -lutil -lrt -lpthread -lm -ldl -lc"
+        );
         assert_eq!(
             static_libraries(libs_msvc, &target_msvc),
             "-lkernel32 -ladvapi32 -lntdll -luserenv -lws2_32 -lmsvcrt"
